@@ -2,6 +2,7 @@ import pathlib
 import tempfile
 import subprocess
 import string
+import os
 
 from cursedspace import ShellContext, InputLine
 
@@ -53,6 +54,23 @@ class OpenItem(Command):
         context.panel.open_selected()
 
 
+@registered_command
+class OpenItemWith(Command):
+    """Open the selected file or folder with a user-specified program"""
+    NAME = 'open-with'
+    ACCEPT_IN = (DocPanel, FilePanel, EditorPanel)
+
+    def completion_options(self, context, *args):
+        if len(args) == 0 or len(args[0]) == 0:
+            return [p.name for p in context.application.known_programs]
+
+        return [p.name for p in context.application.known_programs
+                if args[0] in p.name]
+
+    def execute(self, context, *args):
+        context.panel.open_selected_with(list(args))
+
+
 @registered_command
 class FocusNextPanel(Command):
     """Put the focus on the next panel"""
@@ -200,7 +218,7 @@ class EditMetadataExternally(Command):
             app.error("No store set up to write into")
             return
 
-        if sidecar == file:
+        if sidecar == file or app.metaindexconf.is_sidecar_file(file):
             context.application.error("Cannot edit metadata of a file that's "
                                       "probably a metadata file")
             return
@@ -208,22 +226,23 @@ class EditMetadataExternally(Command):
         if sidecar.is_file():
             if is_collection:
                 meta = store.get_for_collection(sidecar, '')
-                meta = utils.collection_meta_as_writable(meta, sidecar.parent)
+                if file not in meta:
+                    is_collection = False
+                    meta = CacheEntry(file)
+                else:
+                    meta = meta[file]
             else:
-                meta = store.get(sidecar, '')
-                meta.popall(metaindex.shared.IS_RECURSIVE, [])
+                meta = store.get(sidecar)
+                meta.pop(metaindex.shared.IS_RECURSIVE)
 
         else:
             # get the metadata for the selected item as a dict
             results = app.cache.get(file)
             logger.debug(f"Cached metadata for {file}: {results}")
-            meta = multidict.MultiDict()
-
             if len(results) > 0:
-                for key, value in results[0][1].items():
-                    if not key.startswith('extra.'):
-                        continue
-                    meta.add(key.split('.', 1)[1], value)
+                meta = results[0]
+            else:
+                meta = CacheEntry(file)
 
         # resolve the external editor
         editor = context.application.get_text_editor(True)
@@ -244,11 +263,11 @@ class EditMetadataExternally(Command):
             fh.seek(0)
             changed = new_content != original
             if is_collection:
-                updated = store.get_for_collection(fh, '', file.parent)
-                updated = utils.collection_meta_as_writable(updated, file.parent)
+                updated = store.get_for_collection(fh, file.parent)
+                updated = updated[file.parent]
             else:
-                updated = store.get(fh, '')
-                updated.popall(metaindex.shared.IS_RECURSIVE, [])
+                updated = store.get(fh)
+                updated.pop(metaindex.shared.IS_RECURSIVE)
 
         if changed:
             store.store(updated, sidecar)
@@ -348,8 +367,11 @@ class LoadBookmarkNameReader(BookmarkNameReader):
     def __init__(self, application):
         super().__init__(application, 'Bookmark to load: ')
 
-    def bookmark_selected(self, mark):
+    def handle_key(self, key):
         self.app.hide_key_help()
+        super().handle_key(key)
+
+    def bookmark_selected(self, mark):
         self.app.load_bookmark(mark)
 
 
@@ -579,6 +601,36 @@ class InvertSelection(Command):
         context.panel.paint()
 
 
+@simple_command("find")
+def find_command(context, *args):
+    """Find text in the current panel"""
+    if context.panel.is_busy:
+        context.application.error("Panel is currently busy")
+        return
+
+    context.panel.find(" ".join(args))
+
+
+@simple_command("find-next")
+def find_next_command(context):
+    """Find the next occurrence"""
+    if context.panel.is_busy:
+        context.appliaction.error("Panel is currently busy")
+        return
+
+    context.panel.find_next()
+
+
+@simple_command("find-prev")
+def find_previous_command(context):
+    """Find the previous occurrence"""
+    if context.panel.is_busy:
+        context.appliaction.error("Panel is currently busy")
+        return
+
+    context.panel.find_previous()
+
+
 @registered_command
 class RunOCR(Command):
     """Run OCR for this document"""
@@ -609,7 +661,7 @@ class RunOCR(Command):
 
         meta = context.application.cache.get(path, False)
         if len(meta) == 0:
-            meta = multidict.MultiDict()
+            meta = CacheEntry(path)
         else:
             meta = meta[0][1]
 
