@@ -6,6 +6,7 @@ import subprocess
 from metaindex import shared
 from metaindex.ocr import Dummy, TesseractOCR
 from metaindex.humanizer import humanize
+from metaindex.shared import DUBLINCORE_TAGS
 import metaindex.cache
 import metaindex.indexer
 
@@ -94,18 +95,13 @@ class EditorLineCompletion(Completion):
 
 class EditorLine(InputLine):
     def __init__(self, panel, text=None, suggestions=None):
-        y, x, _, _ = panel.content_area()
-        y += panel.pos[0] + panel.cursor - panel.offset
-        x += panel.pos[1] + panel.columns[0]
         if text is None:
             text = panel.app.as_printable(panel.selected_line.value, None)
         super().__init__(panel.app,
                          panel.columns[1],
-                         (y, x),
+                         (1, 1),
                          text=text,
                          background='░')
-        if len(suggestions) > 0:
-            self.completion = EditorLineCompletion(self, suggestions)
 
         self.item = panel.selected_line
         self.parent = panel
@@ -113,8 +109,19 @@ class EditorLine(InputLine):
         self.app.current_panel = self
         self.original_text = self.text
 
+        self.reposition()
+
+        if suggestions is not None and len(suggestions) > 0:
+            self.completion = EditorLineCompletion(self, suggestions)
+
         logger.debug(f"Enter tag edit mode for {self.item} (text: '{text}')")
 
+    def reposition(self):
+        y, x, _, _ = self.parent.content_area()
+        y += self.parent.pos[0] + self.parent.cursor - self.parent.offset
+        x += self.parent.pos[1] + self.parent.columns[0]
+        self.move(y, x)
+
     def handle_key(self, key):
         if key in [Key.TAB, "^N"] and \
            (self.completion is None or not self.completion.is_visible):
@@ -157,6 +164,9 @@ class EditorPanel(ListPanel):
 
     CONFIG_ICON_MULTILINE = 'multiline-indicator'
     CONFIG_ICON_CUTOFF = 'cutoff-indicator'
+    CONFIG_NO_COMPLETION = 'no-completion'
+    CONFIG_TAGS = 'tags'
+    CONFIG_TAGS_DEFAULT = ','.join({'*'} | DUBLINCORE_TAGS)
 
     def __init__(self, filepath, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -176,9 +186,14 @@ class EditorPanel(ListPanel):
 
         self._multiline_icon = ' '
         self._cutoff_icon = ' '
+        self.suggested_tags = set()
+        self._no_completion = set()
         self.configuration_changed()
 
         self.reload()
+        # signal to self.paint() that we have to recalculate the columns,
+        # as the panel will not have the correct size just now
+        self.columns = [0, 0]
         self.cursor = 1
 
     @property
@@ -211,10 +226,11 @@ class EditorPanel(ListPanel):
     def on_copy(self, item):
         return clipboard.ClipboardItem(item, self)
 
-    def on_paste(self, items):
+    def on_paste(self, items, behaviour):
         if self.is_busy:
             raise RuntimeError("Cannot paste right now, editor is busy")
 
+        # TODO - make use of the 'behaviour' flag
         changes = [Insert(shared.EXTRA[:-1], item.data.tag, item.data.value)
                    for item in items if isinstance(item.data, Line)]
         if len(changes) == 0:
@@ -245,6 +261,8 @@ class EditorPanel(ListPanel):
         self.app.paint_focus_bar()
 
     def paint(self, clear=False):
+        if self.columns == [0, 0]:
+            self.rebuild_columns()
         super().paint(clear)
 
         if self.change_ptr > 0 and (self.border & self.BORDER_TOP) > 0:
@@ -310,6 +328,14 @@ class EditorPanel(ListPanel):
            (text is None and '\n' in str(self.selected_line.value.raw_value)):
             self.multiline_edit()
         else:
+            self._create_editor(text)
+            self.app.paint(True)
+
+    def _create_editor(self, text):
+        if self.selected_line is None:
+            return
+        words = None
+        if self.selected_line.tag not in self._no_completion:
             tag = self.selected_line.prefix + "." + self.selected_line.tag
             words = set(self.app.configuration.list(DICTIONARY_SCOPE,
                                                     self.selected_line.tag,
@@ -323,10 +349,9 @@ class EditorPanel(ListPanel):
                 words |= set(sum([[str(v.raw_value) for v in result[tag]]
                                   for result in self.app.cache.find(tag + "?")],
                                  start=[]))
-            self.editor = EditorLine(self,
-                                     text=text,
-                                     suggestions=words)
-            self.app.paint(True)
+        self.editor = EditorLine(self,
+                                 text=text,
+                                 suggestions=words)
 
     def cancel_edit(self):
         if self.editor is None:
@@ -339,6 +364,14 @@ class EditorPanel(ListPanel):
     def resize(self, *args):
         super().resize(*args)
         self.rebuild_columns()
+        self.scroll()
+        if self.editor is not None:
+            self.editor.resize(self.columns[1])
+
+    def move(self, *args):
+        super().move(*args)
+        if self.editor is not None:
+            self.editor.reposition()
 
     def reset(self):
         self.changes = []
@@ -376,8 +409,6 @@ class EditorPanel(ListPanel):
             if is_collection:
                 collection_extra = store.get_for_collection(sidecar_file)
 
-                # collection_extra = \
-                #     utils.collection_meta_as_writable(collection_extra, sidecar_file.parent)
                 logger.debug("Read collection metadata: %s", collection_extra)
 
                 if self.selected_path in collection_extra:
@@ -478,12 +509,16 @@ class EditorPanel(ListPanel):
 
     def rebuild_columns(self):
         if len(self.items) > 1:
-            self.columns = [max([1 if isinstance(row, Header) else len(str(row.tag))+self.SPACING
-                                 for row in self.items]), 0]
-            self.columns[1] = self.dim[1] - self.columns[0] - 2
+            labelwidth = max([len(str(row.tag)) + self.SPACING
+                              for row in self.items
+                              if not isinstance(row, Header)])
+            labelwidth = min(labelwidth, self.dim[1]//3)
+            self.columns = [labelwidth,
+                            self.dim[1] - labelwidth - 2]
         else:
-            half_w = self.dim[1]//2
-            self.columns = [half_w, self.dim[1] - half_w]
+            labelwidth = self.dim[1]//3
+            self.columns = [labelwidth, self.dim[1] - labelwidth - 2]
+        logger.debug(f"maxwidth: {self.dim[1]}, columns: {self.columns}")
 
     def rebuild_items(self):
         self.items = []
@@ -532,12 +567,15 @@ class EditorPanel(ListPanel):
 
     def do_paint_item(self, y, x, maxwidth, is_selected, item):
         if isinstance(item, Header):
-            self.win.addstr(y, x, " "*maxwidth)
-            self.win.addstr(y, x, item.title[:self.dim[1]-x-2], curses.A_BOLD)
+            self.win.addstr(y, x, " "*(self.content_area()[3]-1))
+            self.win.addstr(y, x, item.title[:self.columns[0]], curses.A_BOLD)
         else:
             for colidx, text in enumerate([item.tag, item.value]):
                 self.win.addstr(y, x, " "*self.columns[colidx])
-                maxlen = self.dim[1]-x-2
+                # maxlen = self.dim[1]-x-2
+                maxlen = self.columns[colidx]
+                if colidx == 0:
+                    maxlen -= self.SPACING
 
                 if text is None:
                     text = ''
@@ -571,15 +609,31 @@ class EditorPanel(ListPanel):
         changed = False
 
         if name is None or name == self.CONFIG_ICON_MULTILINE:
-            new_value = self.app.configuration.get(self.SCOPE, self.CONFIG_ICON_MULTILINE, '…')
+            new_value = self.app.configuration.get(self.SCOPE,
+                                                   self.CONFIG_ICON_MULTILINE,
+                                                   '…')
             changed = self._multiline_icon != new_value
             self._multiline_icon = new_value
 
         if name is None or name == self.CONFIG_ICON_CUTOFF:
-            new_value = self.app.configuration.get(self.SCOPE, self.CONFIG_ICON_CUTOFF, '→')
+            new_value = self.app.configuration.get(self.SCOPE,
+                                                   self.CONFIG_ICON_CUTOFF,
+                                                   '→')
             changed = self._multiline_icon != new_value
             self._cutoff_icon = new_value
 
+        if name is None or name == self.CONFIG_NO_COMPLETION:
+            new_value = self.app.configuration.list(self.SCOPE,
+                                                    self.CONFIG_NO_COMPLETION,
+                                                    'title')
+            self._no_completion = set(new_value)
+
+        if name is None or name == self.CONFIG_TAGS:
+            new_values = set(self.app.configuration.list(self.SCOPE,
+                                                         self.CONFIG_TAGS,
+                                                         self.CONFIG_TAGS_DEFAULT))
+            self.suggested_tags = new_values
+
         if changed:
             if self.win is not None:
                 self.scroll()
@@ -694,10 +748,12 @@ class AddTagCommand(command.Command):
 
     def completion_options(self, context, *args):
         text = "" if len(args) == 0 else args[0]
-        keys = {key.split('.', 1)[1]
-                for key in context.application.cache.keys()
-                if key.startswith(shared.EXTRA)}
-        keys |= shared.DUBLINCORE_TAGS
+        keys = context.application.previous_focus.suggested_tags.copy()
+        if '*' in keys:
+            keys.remove('*')
+            keys |= {key.split('.', 1)[1]
+                     for key in context.application.cache.keys()
+                     if key.startswith(shared.EXTRA)}
         return list(key for key in sorted(keys) if key.startswith(text))
 
     def execute(self, context, name=None):
