@@ -15,10 +15,9 @@ try:
 except ImportError:
     grp = None
 
-from multidict import MultiDict
-
 from cursedspace import Panel, colors, ShellContext, Completion
 
+import metaindex.shared
 from metaindex import stores
 from metaindex.humanizer import format_filesize, format_datetime
 from metaindex.configuration import SECTION_GENERAL, CONFIG_COLLECTION_METADATA
@@ -26,6 +25,7 @@ from metaindex.configuration import SECTION_GENERAL, CONFIG_COLLECTION_METADATA
 from metaindexmanager import command
 from metaindexmanager import utils
 from metaindexmanager import clipboard
+from metaindexmanager.clipboard import PasteBehaviour
 from metaindexmanager.utils import logger
 from metaindexmanager.panel import ListPanel, register
 
@@ -117,12 +117,31 @@ class FolderObserver(threading.Thread):
         self.join()
 
 
+class CopyOperation:
+    def __init__(self, blocker):
+        self.blocker = blocker
+        self.counter = 0
+
+    def cancel_requested(self):
+        return self.blocker is not None and \
+               self.blocker.cancel_requested
+
+    def copy(self, src, dst, *args):
+        if self.cancel_requested():
+            return None
+        retval = shutil.copy2(src, dst, *args, follow_symlinks=False)
+        self.counter += 1
+        return retval
+
+
 @register
 class FilePanel(ListPanel):
     SCOPE = 'files'
 
     CONFIG_ICONS = 'use-icons'
     CONFIG_SELECT = 'selection-icon'
+    CONFIG_SHOW_HIDDEN = 'show-hidden-files'
+    CONFIG_SHOW_SIDECARS = 'show-sidecar-files'
     CONFIG_INFO = 'info'
 
     INFO_COLUMNS = {'size', 'bytes', 'owner', 'group',
@@ -137,6 +156,7 @@ class FilePanel(ListPanel):
         self.previous_path = None
         self.post_load = None
         self.show_hidden_files = False
+        self.show_sidecar_files = True
         self.fancy_icons = False
         self.observer = None
         self.cursor_cache = {}
@@ -157,6 +177,7 @@ class FilePanel(ListPanel):
         super().configuration_changed(name)
 
         changed = False
+        must_reload = False
 
         if name is None or name == self.CONFIG_ICONS:
             value = self.app.configuration.bool(self.SCOPE, self.CONFIG_ICONS, 'no')
@@ -181,8 +202,26 @@ class FilePanel(ListPanel):
             if changed:
                 self.calculate_info_columns()
 
+        if name is None or name == self.CONFIG_SHOW_HIDDEN:
+            value = self.app.configuration.bool(self.SCOPE,
+                                                self.CONFIG_SHOW_HIDDEN,
+                                                'no')
+            changed = value != self.show_hidden_files
+            must_reload = changed
+            self.show_hidden_files = value
+
+        if name is None or name == self.CONFIG_SHOW_SIDECARS:
+            value = self.app.configuration.bool(self.SCOPE,
+                                                self.CONFIG_SHOW_SIDECARS,
+                                                'yes')
+            changed = value != self.show_sidecar_files
+            must_reload = changed
+            self.show_sidecar_files = value
+
         if changed:
             if self.win is not None:
+                if must_reload:
+                    self.reload()
                 self.scroll()
                 self.paint(True)
 
@@ -205,8 +244,10 @@ class FilePanel(ListPanel):
             self.observer.start()
 
     def finish_change_path(self, items):
+        is_sidecar = self.app.metaindexconf.is_sidecar_file
         self.items = [i.copy() for i in sorted(items, key=lambda f: [not f.is_dir, f.name])
-                      if self.show_hidden_files or not utils.is_hidden(i)]
+                      if (self.show_hidden_files or not utils.is_hidden(i)) and
+                         (self.show_sidecar_files or not is_sidecar(i.path))]
 
         if self.path in self.cursor_cache:
             self.cursor = self.cursor_cache[self.path]
@@ -292,7 +333,7 @@ class FilePanel(ListPanel):
             raise RuntimeError(f"File '{item.fullpath}' can not be read")
         return clipboard.ClipboardItem(item, self)
 
-    def on_paste(self, items):
+    def on_paste(self, items, behaviour):
         if self.is_busy:
             raise RuntimeError("Cannot paste. Busy")
 
@@ -309,44 +350,71 @@ class FilePanel(ListPanel):
                 raise RuntimeError(f"Cannot paste a {type(item.data).__name__}")
             paths.append((item, is_cut))
 
-        self.run_blocking(self.do_paste, paths)
+        self.run_blocking(self.do_paste, paths, behaviour)
+
+    def do_paste(self, blocker, paths, behaviour):
+        copier = CopyOperation(blocker)
 
-    def do_paste(self, blocker, paths):
         if blocker is not None:
-            blocker.title("Copying")
+            if all(cut for _, cut in paths):
+                blocker.title("Moving")
+            else:
+                blocker.title("Copying")
 
         # copy the files here
         copied_files = []
         copied_directories = []
         moved_files = []
         moved_directories = []
+        cut_files = set()
         for path, is_cut in paths:
-            extension = ''
-            while True:
-                new_fname = path.stem + extension + path.suffix
-                if not (self.path / new_fname).exists():
-                    break
-                if len(extension) == 0:
-                    extension = '_1'
-                else:
-                    extension = '_' + str(int(extension[1:])+1)
+            if copier.cancel_requested():
+                break
+
+            new_fname = path.name
+            if (self.path / new_fname).exists() and behaviour == PasteBehaviour.ERROR:
+                self.app.error(f"File with the same name {new_fname} exists at destination.")
+                continue
+
+            if behaviour == PasteBehaviour.APPEND:
+                # find a good name that doesn't clash with existing files
+                # unless 'overwrite' is True
+                extension = ''
+                while True:
+                    new_fname = path.stem + extension + path.suffix
+                    if not (self.path / new_fname).exists():
+                        break
+                    if len(extension) == 0:
+                        extension = '_1'
+                    else:
+                        extension = '_' + str(int(extension[1:])+1)
+
+            elif behaviour not in [PasteBehaviour.OVERWRITE, PasteBehaviour.ERROR]:
+                raise RuntimeError(f"Unknown paste behaviour {str(behaviour)}")
+
             new_path = self.path / new_fname
+
             if is_cut:
                 # move
                 try:
                     is_dir = path.is_dir()
-                    shutil.move(path, new_path)
+                    shutil.move(path, new_path, copy_function=copier.copy)
                     if is_dir:
                         moved_directories.append((path, new_path))
                     elif path.stem not in self.metaindexfilenames:
+                        cut_files.add(path)
                         moved_files.append((path, new_path))
                 except OSError as exc:
                     self.app.error(f"Failed to move {path.name}: {exc}")
+
             else:
                 # only copy
                 try:
                     if path.is_dir():
-                        shutil.copytree(path, new_path)
+                        shutil.copytree(path, new_path,
+                                        symlinks=True,
+                                        dirs_exist_ok=True,
+                                        copy_function=copier.copy)
                         copied_directories.append(new_path)
                     elif path.stem:
                         shutil.copy2(path, new_path)
@@ -355,97 +423,122 @@ class FilePanel(ListPanel):
                 except OSError as exc:
                     self.app.error(f"Failed to copy {path.name}: {exc}")
 
-        # TODO - do this really by copying the information in the database
-        # self.app.cache.refresh(copied_directories)
         logger.debug("Copied files: %s", copied_files)
 
-        preferred_stores = self.app.metaindexconf.get_preferred_sidecar_stores()
+        # just a shortcut for convenience
+        find_all_sidecars = self.app.metaindexconf.find_all_sidecar_files
 
-        # obtain any cached 'extra' metadata and write it into sidecar files
+        # keep all inserted operations around, to update the cache once
+        # everything's done
         inserts = []
+
         # in case the transfer of metadata/sidecar files fails, make sure any
         # deletions are *not* carried out
-        prevent_deletions = False
+        prevent_deletions = set()
 
+        # collection metadata files that must be updated
+        touched_collections = {}
+
+        # direct sidecar files that should be deleted
+        sidecars_to_delete = set()
+
+        # handle the sidecar files of all copied and moved files
         for old_path, new_path in copied_files + moved_files:
-            entries = self.app.cache.get(old_path)
-            if len(entries) == 0:
+            last_modified = metaindex.shared.get_last_modified(new_path)
+            merged_metadata = metaindex.shared.CacheEntry(new_path,
+                                                          last_modified=last_modified)
+
+            sidecars = list(find_all_sidecars(new_path))
+            if len(sidecars) > 0:
+                if behaviour == PasteBehaviour.ERROR:
+                    prevent_deletions.add(old_path)
+                    self.app.error(f"Will not overwrite existing sidecar file at {new_path}")
+                    continue
+                if behaviour == PasteBehaviour.OVERWRITE:
+                    sidecars = []
+                # APPEND behaviour is automatic, as the existing sidecar will be considered
+                # when copying the "old" metadata
+
+            # find all sidecar files and merge their data for this old_path
+            # file into one sidecar file
+            sidecars += find_all_sidecars(old_path)
+            for sidecar, is_collection in sidecars:
+                if is_collection:
+                    metadata = stores.get_for_collection(sidecar)
+                    if old_path.parent in metadata:
+                        merged_metadata.update(metadata[old_path.parent])
+                    if old_path in metadata:
+                        merged_metadata.update(metadata[old_path])
+                        if old_path in cut_files and sidecar not in touched_collections:
+                            touched_collections[sidecar] = metadata
+                else:
+                    metadata = stores.get(sidecar)
+                    merged_metadata.update(metadata)
+                    if old_path in cut_files:
+                        sidecars_to_delete.add(sidecar)
+
+            # make sure we're keeping this!
+            # in case the target sidecar file existed already, merged_metadata
+            # might have picked up a newer 'last_modified' from it
+            merged_metadata.last_modified = last_modified
+
+            if len(merged_metadata) == 0:
+                # if there's no metadata to store, we are out
                 continue
-            metadata = entries[0].metadata
-            last_modified = entries[0].last_modified
-            inserts.append((new_path, metadata, last_modified))
-
-            # get all extra metadata, but strip the 'extra.' prefix
-            writeable_metadata = MultiDict([(k.split('.', 1)[1], v)
-                                            for k, v in metadata.items()
-                                            if k.startswith('extra.')])
-
-            # if there is a direct sidecar file, the user probably wants
-            # a direct sidecar file again for the new file
-            direct_sidecars = [fn
-                               for fn, is_collection in
-                                   self.app.metaindexconf.find_all_sidecar_files(old_path)
-                               if not is_collection]
-
-            # see if there is any other sidecar file in place already
-            sidecar, is_collection, store = \
-                self.app.metaindexconf.resolve_sidecar_for(new_path)
 
+            sidecar, _, store = self.app.metaindexconf.resolve_sidecar_for(new_path)
             logger.debug("trying to put %s metadata tags into %s",
-                         len(writeable_metadata), sidecar)
+                         len(merged_metadata), sidecar)
 
-            if len(direct_sidecars) > 0:
-                for sidecar in direct_sidecars:
-                    new_sidecar = new_path.parent / (new_path.stem + sidecar.suffix)
-                    try:
-                        shutil.copy2(sidecar, new_sidecar)
-                    except OSError as exc:
-                        logger.error("Failed to copy %s to %s: %s",
-                                     sidecar, new_sidecar, exc)
-                        self.app.error("Could not copy all sidecar files. See logfile for details")
-                        prevent_deletions = True
+            logger.debug("writing into direct sidecar %s: %s",
+                         sidecar, merged_metadata)
+            try:
+                store.store(merged_metadata, sidecar)
+            except (OSError, IOError, ValueError) as exc:
+                logger.error("Could not write extra metadata of "
+                             "%s (copy of %s): %s",
+                             new_path, old_path, exc)
+                prevent_deletions.add(old_path)
 
-            elif sidecar.exists():
-                assert is_collection
-                existing = store.get_for_collection(sidecar, '')
-                existing = utils.collection_meta_as_writable(existing, new_path.parent)
-                existing[str(new_path.name)] = writeable_metadata
-                logger.debug("Writing collection %s into %s",
-                             existing, sidecar)
-                store.store(existing, sidecar)
-
-            elif len(preferred_stores) > 0:
-                store = preferred_stores[0]
-                sidecar = new_path.parent / (new_path.stem + store.SUFFIX)
-                logger.debug("writing into direct sidecar %s: %s", sidecar, writeable_metadata)
-                store.store(writeable_metadata, sidecar)
-
-            else:
-                logger.error("Could not write extra metadata of %s (copy of %s): "
-                             "no available stores",
-                             new_path, old_path)
-                prevent_deletions = True
+            # remember to write the new data into the cache
+            inserts.append(merged_metadata)
 
+        # write all metadata of pasted files into the cache
         if len(inserts) > 0:
             self.app.cache.bulk_insert(inserts)
 
         if len(moved_directories) > 0:
             self.app.cache.bulk_rename([(old, new, True) for old, new in moved_directories])
 
+        # now delete old sidecar files
         deletes = []
-        if not prevent_deletions:
-            for old_path, _ in moved_files:
-                for sidecar, is_collection in self.app.metaindexconf.find_all_sidecar_files(old_path):
-                    if is_collection:
-                        # TODO - remove the file-specific entries from the sidecar file
-                        continue
-                    try:
-                        sidecar.unlink()
-                    except OSError as exc:
-                        self.app.error(f"Failed to delete sidecar file '{sidecar.name}': {exc}")
-                deletes.append(old_path)
+        for old_path, _ in moved_files:
+            # remember to delete the old path from the cache
+            deletes.append(old_path)
 
-        if not prevent_deletions and len(deletes) > 0:
+            if old_path in prevent_deletions:
+                # something went wrong when creating the sidecar file for this
+                # old file, so don't touch/delete the old sidecar(s)
+                continue
+
+            for sidecar, metadata in touched_collections.items():
+                if old_path in metadata:
+                    del metadata[old_path]
+
+        for sidecar, metadata in touched_collections.items():
+            assert isinstance(metadata, dict)
+            try:
+                stores.store(list(metadata.values()), sidecar)
+            except OSError as exc:
+                self.app.error(f"Failed to update '{sidecar.name}': {exc}")
+
+        for sidecar in sidecars_to_delete:
+            try:
+                sidecar.unlink()
+            except OSError as exc:
+                self.app.error(f"Failed to delete sidecar file '{sidecar.name}': {exc}")
+
+        if len(deletes) > 0:
             self.app.cache.forget(deletes)
 
         if self.observer is not None:
@@ -850,7 +943,23 @@ class ToggleShowHiddenFiles(command.Command):
         if context.panel.is_busy:
             return
         context.panel.show_hidden_files = not context.panel.show_hidden_files
-        context.panel.change_path(context.panel.path)
+        context.panel.reload()
+        context.panel.scroll()
+        context.panel.paint(True)
+
+
+@command.registered_command
+class ToggleShowSidecarFiles(command.Command):
+    """Toggle show sidecar files"""
+    NAME = 'toggle-sidecar'
+    ACCEPT_IN = (FilePanel,)
+
+    def execute(self, context):
+        if context.panel.is_busy:
+            return
+        context.panel.show_sidecar_files = not context.panel.show_sidecar_files
+        context.panel.reload()
+        context.panel.scroll()
         context.panel.paint(True)
 
 
@@ -902,7 +1011,7 @@ def shell_command(context, *args):
     scr = context.application.screen
     with ShellContext(scr, True):
         try:
-            subprocess.run(args, shell=False, check=False)
+            subprocess.run(" ".join(args), shell=True, check=False)
         except (KeyboardInterrupt,):
             pass
 
@@ -912,3 +1021,27 @@ def shell_command(context, *args):
             scr.get_wch()
             scr.timeout(context.application.key_timeout)
     context.application.paint(True)
+
+
+@command.simple_command('launch', accept_in=(FilePanel,))
+def launch_command(context, *args):
+    """Launch a command here"""
+
+    if len(args) == 0:
+        shell = shutil.which(os.getenv('SHELL') or "")
+        if shell is None:
+            context.application.error("No SHELL found")
+            return
+        args = [shell]
+
+    logger.debug(f"Launching in shell: {args}")
+
+    try:
+        subprocess.Popen(" ".join(args),
+                         shell=True,
+                         bufsize=0,
+                         stdin=subprocess.DEVNULL,
+                         stdout=subprocess.DEVNULL,
+                         stderr=subprocess.DEVNULL)
+    except OSError:
+        pass
