@@ -154,6 +154,8 @@ class Application(cursedspace.Application):
             (FilePanel.SCOPE, ('y', 'a'), ('append',)),
             (FilePanel.SCOPE, ('d', 'a'), ('cut-append',)),
             (FilePanel.SCOPE, ('p', 'p'), ('paste',)),
+            (FilePanel.SCOPE, ('p', 'o'), ('paste-overwrite',)),
+            (FilePanel.SCOPE, ('p', 'a'), ('paste-append',)),
             (FilePanel.SCOPE, ('!', ), (':shell',)),
             (DocPanel.SCOPE, ('y', 'y'), ('copy',)),
             (DocPanel.SCOPE, ('y', 'a'), ('append',)),
@@ -162,6 +164,7 @@ class Application(cursedspace.Application):
             (EditorPanel.SCOPE, ('p', 'p'), ('paste',)),
             (shared.ANY_SCOPE, ('y', '0'), ('clear-clipboard',)),
             (FilePanel.SCOPE, ('z', 'h'), ('toggle-hidden',)),
+            (FilePanel.SCOPE, ('z', 'm'), ('toggle-sidecar',)),
             (shared.ANY_SCOPE, ('z', 's'), ('::layout horizontal', 'Split layout')),
             (shared.ANY_SCOPE, ('z', 't'), ('::layout tabbed', 'Tabbed layout')),
             (shared.ANY_SCOPE, (':', ), ('enter-command',)),
@@ -216,7 +219,12 @@ class Application(cursedspace.Application):
         conf = configparser.ConfigParser(interpolation=None)
         conf.read_dict({shared.ALL_SCOPE: {PREF_OPENER: DEFAULT_OPENER,
                                            PREF_INFO_TIMEOUT: DEFAULT_INFO_TIMEOUT,
-                                           shared.PREF_BORDER: shared.DEFAULT_BORDER}})
+                                           shared.PREF_BORDER: shared.DEFAULT_BORDER},
+                        EditorPanel.SCOPE: {
+                            EditorPanel.CONFIG_NO_COMPLETION: 'title',
+                            EditorPanel.CONFIG_TAGS: EditorPanel.CONFIG_TAGS_DEFAULT,
+                            },
+                        })
         self.configuration = MetaIndexManagerConfiguration(conf)
         utils.parse_ls_colors()
         utils.parse_ls_icons()
@@ -271,8 +279,12 @@ class Application(cursedspace.Application):
         self.known_programs = []
         self.refresh_known_programs()
 
-        configpath = pathlib.Path(args.config).expanduser().resolve()
-        self.load_config_file(configpath)
+        configpath = pathlib.Path(args.config).expanduser()
+        if configpath.parent.exists():
+            here = os.getcwd()
+            os.chdir(configpath.parent)
+            self.load_config_file(configpath)
+            os.chdir(here)
 
     def main(self):
         self.command_input = None
@@ -318,6 +330,9 @@ class Application(cursedspace.Application):
                         pass
                     redraw = True
 
+                    if redraw and self.blocking_task is not None:
+                        self.blocking_task.paint()
+
             if not timed_out or redraw:
                 if self.key_help_panel is not None:
                     self.key_help_panel.paint()
@@ -384,9 +399,13 @@ class Application(cursedspace.Application):
                         if commandname.startswith('::'):
                             args = shlex.split(commandname[2:])
                             commandname = args.pop(0)
+                            args = sum([expand_part(self, part)
+                                        for part in args],
+                                       start=[])
 
                         command = resolve_command(commandname)
-                        if command is not None and isinstance(self.current_panel, command.ACCEPT_IN):
+                        if command is not None and \
+                           isinstance(self.current_panel, command.ACCEPT_IN):
                             if args is None:
                                 self.execute_command(commandname)
                             else:
@@ -1051,6 +1070,7 @@ def run():
         return 0
 
     logger.debug("Starting metaindexmanager")
+    logging.getLogger().addHandler(logger.handler)
 
     if args.log_level.lower() in ['debug']:
         metaindex.logger.setup(level=logging.DEBUG, filename=args.log_file.parent / 'metaindex.log')
