@@ -8,8 +8,7 @@ import pathlib
 import subprocess
 import datetime
 import string
-
-import multidict
+from logging.handlers import RotatingFileHandler
 
 import metaindex.shared
 import metaindex.ocr
@@ -28,7 +27,10 @@ class MetaindexmanagerLogger:
         if filename is None:
             self.handler = logging.NullHandler()
         else:
-            self.handler = logging.FileHandler(filename, encoding='utf-8')
+            self.handler = RotatingFileHandler(filename,
+                                               encoding='utf-8',
+                                               maxBytes=4096,
+                                               backupCount=1)
 
         self.logger = logging.getLogger(PROGRAMNAME)
         self.logger.propagate = False
@@ -386,39 +388,6 @@ def is_hidden(path, stats=None):
     return any((on_windows, on_mac, on_linux))
 
 
-def collection_meta_as_writable(data, basepath):
-    """Change the 'data', as read from stores.get_for_collection(), to be usable for stores.store
-
-    basepath is the parent path of the metadata file.
-    """
-    # turn the collection_extra again into the editable plain form,
-    # i.e. restore '*', '**', and filename entries
-    for key in set(data.keys()):
-        if key == basepath:
-            nkey = '*'
-            if data[key].popall(metaindex.shared.IS_RECURSIVE) == [True]:
-                nkey = '**'
-            data[nkey] = data.pop(key)
-            key = nkey
-        else:
-            nkey = key.name
-            if isinstance(data, multidict.MultiDict):
-                specifics = data.popall(key, [])
-                if len(specifics) > 1:
-                    merged = multidict.MultiDict()
-                    for specific in specifics:
-                        merged.extend(specific)
-                    specifics = merged
-            else:
-                assert isinstance(data, dict)
-                specifics = data.pop(key)
-
-            data[nkey] = specifics
-            key = nkey
-        data[key].pop(metaindex.shared.IS_RECURSIVE)
-    return data
-
-
 def first_line(text):
     """Return only the first line of a potential multi-line text"""
     return text.split("\n")[0].split("\r")[0]
